#import <EOControl/EOFault.h>
#import <Foundation/NSDebug.h>
#import <objc/objc-class.h>

@implementation EOFault (OOAFixes)

typedef struct _EOFaultStruct {
    Class  isa;
    void  *_handler;
} EOFaultStruct;

// This fixes a corruption of objects that have inline retain counts stored in their
// second word (the one after the isa).  This also helps fix the bogus reporting
// of allocation statistics during the transfer of references to the fault.

//
// This patch should ONLY be used with EOF2.1.  Other versions of EOF may or may
// not have this bug or may require that these methods be implemented differently.
//

+ (void)makeObjectIntoFault:(id)eo withHandler:(EOFaultHandler *)handler
{
    EOFaultStruct *rawEo;
    Class          faultClass = (Class)self, targetClass;
    unsigned       retainCount;
    BOOL           oldAllocationFlag;
    
    rawEo = (EOFaultStruct *)eo;
    targetClass = rawEo->isa;
    
    //
    // Check if we have enougth space in the instance
    //
    if (targetClass->instance_size < faultClass->instance_size)
	[NSException raise:NSInvalidArgumentException format:@"*** %s +%s cannot create a fault for a class %s because its instances are smaller than %d bytes", ((Class)self)->isa->name, sel_getName (_cmd), targetClass->name, faultClass->instance_size];

    //
    // Transfert the retainCount from the object to the handler
    // Keep a retainCount of 1 in the object so it's not released.
    // Turn off allocation statistics while transfering references
    // since otherwise we will transmit inconsistent stats
    //

    // We should really protect this with a lock but EOF is already so unthread-safe
    // it doesn't really matter.
    oldAllocationFlag = NSKeepAllocationStatistics;
    NSKeepAllocationStatistics = NO;

    // If this code raises due to an error occuring in -retainCount or -release
    // on the EO, the allocation statistics will be permanently disabled.
    // This doesn't seem like enough of a threat to add an exception handler
    // and take the corresponding performance hit.

    retainCount = [eo retainCount]; // retainCount should be at least 1
    
    while((--retainCount) > 0) { // The > 0 is here only for safety reason.
        [eo release];
        [handler incrementExtraRefCount];
    }

    // Restore allocation stats
    NSKeepAllocationStatistics = oldAllocationFlag;

    //
    // Initialize the handler.  Do not read any values from the guts of the object
    // (ie, the extraData) until we are done messaging it.  Hell, for all we know,
    // the object could change its isa during retain.
    //
    [handler setTargetClass: rawEo->isa
                  extraData: rawEo->_handler];

    //
    // Setup the isa pointer and the handler
    //
    rawEo->isa = faultClass;
    rawEo->_handler = [handler retain];
}


// This override helps work around the bogus reporting of allocation statistics
// during unfaulting.

+ (void)clearFault:(id)fault
{
    EOFaultStruct  *object = (EOFaultStruct *)fault;
    EOFaultHandler *faultHandler = object->_handler;
    unsigned        extraRefCount;
    Class           targetClass;
    void           *data;
    BOOL            oldAllocationFlag;

    //
    // Cannot clear something that's not a fault !
    //
    if (![self isFault:fault])
	[NSException raise:NSInvalidArgumentException format:@"*** %s %s object not a fault", ((Class)self)->isa->name, sel_getName (_cmd)];

    //
    // Notify the fault handler
    //
    [faultHandler faultWillFire:fault];

    //
    // Get all the data from the fault handler
    //
    extraRefCount = [faultHandler extraRefCount];
    targetClass = [faultHandler targetClass];
    data = [faultHandler extraData];

    //
    // Turn the object back to normal
    //
    object->isa = targetClass;
    object->_handler = (EOFaultHandler *)data;

    // Temporarily turn off allocation statistics to avoid reporting the
    // transfer of references.  Like above, we won't lock or put an
    // exception handler around this.
    oldAllocationFlag = NSKeepAllocationStatistics;
    NSKeepAllocationStatistics = NO;

    //
    // Update the retainCount in the object
    //
    // The current refcount in the object should be 1 already
    //
    while (extraRefCount--)
        [fault retain];

    // Restore the stats flag
    NSKeepAllocationStatistics = oldAllocationFlag;

    //
    // Drop the fault handler
    //
    [faultHandler release];
}

@end
