/* OOAInternalRefCountExample.m created by bungi on Fri 14-Nov-1997 */

#import "OOAInternalRefCountExample.h"

// The Rhapsody build process doesn't define this automatically,
// you'll need to uncomment this line on Rhapsody
#warning Comment out the following line if you are building on OpenStep/Mach
#define RHAPSODY

#ifdef RHAPSODY

#import <Foundation/NSDate.h>  // For NSTimeInterval, needed by NSDebug.h
#import <Foundation/NSPort.h>  // For NSPort, needed by NSDebug.h
#import <Foundation/NSDebug.h>

#else

// These are defined in NSObject but NeXT didn't make them fully public.
// (They have been made public in the Rhapsody DR.)  Objects that do
// not call the superclass version of -allocWithZone:, -dealloc,
// -copyWithZone:, -autorelaese, -retain, or -release should call
// NSRecordAllocationEvent.
// See NSDebug.h for more details.
#import <Foundation/NSDebug.h>

typedef enum _NSAllocationEvent {
    NSObjectAllocatedEvent = 0,           // Sent by NSAllocateObject (which is called from +[NSObject allocWithZone:])
    NSObjectDeallocatedEvent,             // Sent by NSDeallocateObject (which is called from -[NSObject dealloc])
    NSObjectCopiedEvent,                  // Sent by NSCopyObject (which is called from -[NSObject copyWithZone:])
    NSObjectAutoreleasedEvent,            // Sent by [NSAutoreleasePool addObject:] (which is called from -[NSObject autorelease])
    NSObjectExtraRefIncrementedEvent,     // Sent by NSIncrementExtraRefCount (which is called from -[NSObject retain])
    NSObjectExtraRefDecrementedEvent,     // Sent by NSDecrementExtraRefCountWasZero (which is called from -[NSObject release])
    NSObjectInternalRefIncrementedEvent,  // Not sent by default
    NSObjectInternalRefDecrementedEvent,  // Not sent by default
    NSObjectPoolDeallocStartedEvent,      // Sent at the start of -[NSAutoreleasePool dealloc]
    NSObjectPoolDeallocFinishedEvent      // Sent at the end of -[NSAutoreleasePool dealloc]
} NSAllocationEvent;

FOUNDATION_EXPORT void NSRecordAllocationEvent(NSAllocationEvent event, id object, void *data1, void *data2);

#endif

//
// This class demonstrates an efficient way to implement the Foundation
// defined reference counting mechanism. The object sets aside some number
// of bits to store the low order bits of the retainCount.  When these
// bits overflow, the 'extra ref count' stored in a Foundation-maintained
// table is incremented.  When the reference count underflows, the external
// table entry is decremented.
//
// If the external reference count is zero, the total reference count for
// the object is one.  This is an optimization designed to avoid requiring
// the object allocation process to insert an entry in the external reference
// count table.
//
// Storing an inline retain count can be an effective optimization when
// the number of reference count changes is large and the object in question
// has some number of bits that would go wasted otherwise.
//
// As an additional optimization, this class maintains a single bit that
// records whether there is currently any external reference count.  If
// the internal reference count never overflows, then by having this
// flag we can avoid ever checking the external table when our reference
// count goes to zero.
//
// Note that this class WILL NOT WORK in a multi-threaded application (in
// the general case).  The reference counting mechanism needs to be thread-safe
// to avoid having two threads modify the reference count at the same time
// producing unpredictable results (either resulting in a leaked object
// or a prematurely deallocated object).
//
// However, if each of the methods below is enclosed by a global mutex, this approach
// works very well (in fact, OmniWeb does exactly this).  One point to note here
// is that this optimization is VERY effective in multi-threaded applications.
// In this case, the external reference table becomes a point of contention
// (since it has its own locks).  By increasing the number of locks in the system
// the contention on each lock is reduced.
//
// In order for OmniObjectAlloc to correctly track allocation events for objects
// that override retainCount/retain/release in this fashion, the object must
// call NSRecordAllocationEvent appropriately.
//

#define OOAInternalRefCountExampleScale  (1 << OOAInternalRefCountExampleWidth)
#define OOAMaxInternalRefCount           (OOAInternalRefCountExampleScale - 1)


@implementation OOAInternalRefCountExample

- (unsigned int) retainCount;
{
    if (info.hasExternalRefCount)
        // Only check the external table if we actually have an external ref count.
        // The external table contains the high order bits of our refCount.
        // Then add in the internal count and the extra implied initial count of one.
        return NSExtraRefCount(self) * OOAInternalRefCountExampleScale + info.internalRefCount + 1;
    else
        // Basically the same formula as above except that we know that
        // NSExtraRefCount(self) is zero.
        return info.internalRefCount + 1;
}


- retain;
{
    if (info.internalRefCount == OOAMaxInternalRefCount) {
        // If our internal ref count field is full, reset it, increment the
        // external table and record that we now have an external entry.
        // We do not need to call NSRecordAllocationEvent in this case
        // since NSIncrementExtraRefCount will do it for us
        info.internalRefCount = 0;
        info.hasExternalRefCount = YES;
        NSIncrementExtraRefCount(self);
    } else {
        // We still have space in our internal field, simply increment it.
        info.internalRefCount++;
        if (NSKeepAllocationStatistics &&
            (NSAllocationStatisticsOutputMask & NSObjectInternalRefIncrementedEventMask)) {
            // Report our allocation statistics to make OOA and 'oh' happy.
            NSRecordAllocationEvent(NSObjectInternalRefIncrementedEvent, self, NULL, NULL);
        }
    }

    return self;
}


- (void) release;
{
    if (!info.internalRefCount) {
        // Our internal table is empty
        if (info.hasExternalRefCount) {
            // We had an external ref count.  Decrement the external count and
            // store whether we still have an external count.  Reset our internal
            // field to the max value.  We do not need to call NSRecordAllocationEvent
            // in this case since NSDecrementExtraRefCountWasZero will do it for us.
            // We compare NSExtraRefCount to one since we want to call
            // NSDecrementExtraRefCountWasZero after having adjusted our internal state.
            // This makes it so that the allocation event recorded has the proper retainCount
            // (yes, it will call -retainCount on us).
            info.hasExternalRefCount = (NSExtraRefCount(self) > 1);
            info.internalRefCount = OOAMaxInternalRefCount;
            NSDecrementExtraRefCountWasZero(self);
        } else
            // We are all out of references.  Time to die.
            [self dealloc];
    } else {
        // We still have internal references.  Simply remove one of those.
        info.internalRefCount--;
        if (NSKeepAllocationStatistics &&
            (NSAllocationStatisticsOutputMask & NSObjectInternalRefDecrementedEventMask)) {
            // Report our allocation statistics to make OOA and 'oh' happy.
            NSRecordAllocationEvent(NSObjectInternalRefDecrementedEvent, self, NULL, NULL);
        }
    }
}


@end

